// ------------------------------------------------------------
// Machinery project - math library
// ------------------------------------------------------------
// (c) 2000 by Sascha 'EvilOne' Salevsky
// E-mail: Sascha.Salevsky@Ibykus.de
// ------------------------------------------------------------

#ifndef _mathematics_h_
#error "Dont include Matrix4 directly!!! Use Mathematics.h instead."
#endif

// Constructors

inline Matrix4::Matrix4()
{}

inline Matrix4::Matrix4(const Matrix4& other)
{
	for(int i = 0; i < 16; i++) M[i] = other.M[i];
}

inline Matrix4::Matrix4(const float* other)
{
	for(int i = 0; i < 16; i++) M[i] = other[i];
}

inline Matrix4::Matrix4(const Quaternion& other)
{
	Assign(other);
}

// Assignment

inline Matrix4& Matrix4::operator=(const Matrix4& other)
{
	for(int i = 0; i < 16; i++) M[i] = other.M[i];
	return *this;
}

inline void Matrix4::Assign(const Matrix4& other)
{
	for(int i = 0; i < 16; i++) M[i] = other.M[i];
}

inline void Matrix4::Assign(const float* other)
{
	for(int i = 0; i < 16; i++) M[i] = other[i];
}

inline Matrix4& Matrix4::operator=(const Quaternion& other)
{
	Assign(other);
	return *this;
}

// Casting

inline Matrix4::operator const float*() const
{
	return M;
}

inline Matrix4::operator float*()
{
	return M;
}

// Simple methods

inline Matrix4& Matrix4::SetIdentity()
{
	for(int i = 0; i < 16; i++) M[i] = Matrix4::Identity_Matrix[i];
	return *this;
}

inline Matrix4& Matrix4::SetZero()
{
	for(int i = 0; i < 16; i++) M[i] = 0.0f;
	return *this;
}

inline Matrix4& Matrix4::SetSwitchOrientation()
{
	for(int i = 0; i < 16; i++) M[i] = Matrix4::Orientation_Switch_Matrix[i];
	return *this;
}

inline Matrix4& Matrix4::SetPerspective()
{
	for(int i = 0; i < 16; i++) M[i] = Matrix4::Perspective_Matrix[i];
	return *this;
}

inline Matrix4& Matrix4::Mul(const Matrix4& other)
{
	Matrix4 tmp;
	tmp.Assign(*this);
	Mul(tmp, other);
	return *this;
}

inline Matrix4& Matrix4::Mul(const Matrix4& m1, const Matrix4& m2)
{
	const float *b = m1.M;
	const float *a = m2.M;
	float *prod = M;

	for(int i = 0; i < 4; i++)
	{
		float a0 = a[i];
		float a1 = a[i + 4];
		float a2 = a[i + 8];
		float a3 = a[i + 12];
		prod[i] = a0 * b[0] + a1 * b[1] + a2 * b[2] + a3 * b[3];
		prod[i + 4] = a0 * b[4] + a1 * b[5] + a2 * b[6] + a3 * b[7];
		prod[i + 8] = a0 * b[8] + a1 * b[9] + a2 * b[10] + a3 * b[11];
		prod[i + 12] = a0 * b[12] + a1 * b[13] + a2 * b[14] + a3 * b[15];
	}
	return *this;
}

inline Matrix4& Matrix4::Mul(float f)
{
	for(int i = 0; i < 16; i++) M[i] *=f;
	return *this;
}

inline Matrix4& Matrix4::Mul(const Matrix4& other, float f)
{
	const float* m = other.M;
	for(int i = 0; i < 16; i++) M[i] = m[i] * f;
	return *this;
}

inline Matrix4& Matrix4::Div(float f)
{
	f = 1.0f / f;
	for(int i = 0; i < 16; i++) M[i] *=f;
	return *this;
}

inline Matrix4& Matrix4::Div(const Matrix4& other, float f)
{
	f = 1.0f / f;
	const float* m = other.M;
	for(int i = 0; i < 16; i++) M[i] = m[i] * f;
	return *this;
}

// Operators

inline Matrix4& Matrix4::operator*=(const Matrix4& other)
{
	Matrix4 tmp;
	tmp.Assign(*this);
	Mul(tmp, other);
	return *this;
}

inline Matrix4 Matrix4::operator*(const Matrix4& other) const
{
	Matrix4 tmp;
	tmp.Mul(*this, other);
	return tmp;
}

// Transformation

inline Matrix4& Matrix4::Translate(float tx, float ty, float tz)
{
	M[12] = M[0] * tx + M[4] * ty + M[8]  * tz + M[12];
	M[13] = M[1] * tx + M[5] * ty + M[9]  * tz + M[13];
	M[14] = M[2] * tx + M[6] * ty + M[10] * tz + M[14];
	M[15] = M[3] * tx + M[7] * ty + M[11] * tz + M[15];
	return *this;
}

inline Matrix4& Matrix4::SetTranslate(float tx, float ty, float tz)
{
	SetIdentity();
	return Translate(tx, ty, tz);
}

inline Matrix4& Matrix4::Scale(float sx, float sy, float sz)
{
	M[0] *= sx; M[4] *= sy; M[8]  *= sz;
	M[1] *= sx; M[5] *= sy; M[9]  *= sz;
	M[2] *= sx; M[6] *= sy; M[10] *= sz;
	M[3] *= sx; M[7] *= sy; M[11] *= sz;
	return *this;
}

inline Matrix4& Matrix4::SetScale(float sx, float sy, float sz)
{
	SetIdentity();
	return Scale(sx, sy, sz);
}

inline Matrix4& Matrix4::Rotate(float rx, float ry, float rz)
{
	Matrix4 tmp;
	tmp.SetRotate(rx, ry, rz);
	Mul(tmp);
	return *this;
}

inline Matrix4& Matrix4::Rotate(float angle, float x, float y, float z)
{
	Matrix4 tmp;
	tmp.SetRotate(angle, x, y, z);
	Mul(tmp);
	return *this;
}

inline Matrix4& Matrix4::SetRotate(const Vector3& r)
{
	return SetRotate(r.x, r.y, r.z);
}

inline Matrix4& Matrix4::Rotate(const Vector3& r)
{
	return Rotate(r.x, r.y, r.z);
}

inline Matrix4& Matrix4::SetRotate(float angle, const Vector3& r)
{
	return SetRotate(angle, r.x, r.y, r.z);
}

inline Matrix4& Matrix4::Rotate(float angle, const Vector3& r)
{
	return Rotate(angle, r.x, r.y, r.z);
}
